/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

#include "Common.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
//#include <ucontext.h>
//#include <execinfo.h>
//#include <signal.h> 

static int log_level = WARN;

// Set the logging level (from DUMP to ERROR)
void
set_log_level (int lvl)
{
    log_level = lvl;
}


// Get the logging level
int
get_log_level ()
{
    return log_level;
}


// Return the name of a debug level
const char* lvl2str (int lvl)
{
    switch (lvl)
    {
        case FATAL: return STR_FATAL;
        case ERROR: return STR_ERROR;
        case WARN:  return STR_WARN;
        case INFO:  return STR_INFO;
        case DUMP:  return STR_DUMP;
        default:    return STR_UNKNOW;
    }
}


#define MSGSIZE_MAX 1024
// Print a log message
void
log_c (int lvl, const char* file, int line, const char* format, ...)
{
    va_list va;
    char tmp[MSGSIZE_MAX];

    //snprintf(tmp, MSGSIZE_MAX, "[%s](%s, %d) ", lvl2str(lvl), file, line);
    struct timeval now;
    gettimeofday (&now, NULL);
    snprintf(tmp, MSGSIZE_MAX, "%u [%s](%s, %d) ", now.tv_sec, lvl2str(lvl), file, line);

    tmp[MSGSIZE_MAX-2] = '\n';
    tmp[MSGSIZE_MAX-1] = '\0';

    int length = strlen(tmp);
    va_start(va, format);
    vsnprintf(tmp + length, MSGSIZE_MAX - length, format, va);
    va_end(va);
    tmp[MSGSIZE_MAX-2] = '\n';
    tmp[MSGSIZE_MAX-1] = '\0';

    if (lvl <= ERROR)
    {
        fprintf(stderr, "%s\n", tmp);
        fflush(stderr);
    }
    else
    {
        fprintf(stderr, "%s\n", tmp);
        fflush(stderr);
    }
}


char str_log_msg[MSGSIZE_MAX];

// Forge a log message
const char*
str_log_c (int lvl, const char* file, int line, const char* format, ...)
{
    va_list va;

    snprintf(str_log_msg, MSGSIZE_MAX, "[%s](%s, %d)", lvl2str(lvl), file, line);
    str_log_msg[MSGSIZE_MAX-2] = '\n';
    str_log_msg[MSGSIZE_MAX-1] = '\0';

    int length = strlen(str_log_msg);
    va_start(va, format);
    vsnprintf(str_log_msg + length, MSGSIZE_MAX - length, format, va);
    va_end(va);
    str_log_msg[MSGSIZE_MAX-2] = '\n';

    return str_log_msg;
}


// Dump a char array in an hexadecimal notation
void
dumpRawData (int lvl, uint8* data, int len)
{
    if (lvl <= log_level)
    {
        int i = 0;
        while (i < len)
        {
            if (lvl <= ERROR)
            {
                if ((i % 8) == 0) fprintf(stderr, "[%s]", lvl2str(lvl));
                fprintf(stderr, "0x%02x ", data[i]);
                if ((i % 8) == 7) fprintf(stderr, "\n");
                fflush(stderr);
            }
            else
            {
                if ((i % 8) == 0) fprintf(stdout, "[%s]", lvl2str(lvl));
                fprintf(stdout, "0x%02x ", data[i]);
                if ((i % 8) == 7) fprintf(stdout, "\n");
                fflush(stdout);
            }
            i++;
        }
        if (lvl <= ERROR) fprintf(stderr, "\n");
        else fprintf(stdout, "\n");
    }
}


void ASSERT(int cond, const char * errformat, ...)
{
    va_list ap;
    if (! cond) {
        fprintf(stderr, "assertion violation: ");
        va_start(ap, errformat);
        vfprintf(stderr, errformat, ap);
        va_end(ap);
        fflush(stderr);
        exit(1);
    }
}

// static void print_trace(void)
// {
//     void *array[50];
//     size_t size;
//     char **strings;
//     size_t i;

//     size = backtrace(array, 50);
//     strings = backtrace_symbols(array, size);

//     fprintf(stderr, "LOCATION: Print call stack\n");

//     for(i = 0; i < size; i++)
//         fprintf(stderr, "%s\n", strings[i]);

//     free(strings);
// }

// void handler(int sig, siginfo_t *sip, void *arg3)
// {
//     printf ("SEG FAULT!!!!\n");
//     print_trace ();
//     exit(1);
// }

// void setup_sigsegv(void)
// {
//     struct sigaction sa;
//     sigset_t signalmask;
    
//     sa.sa_sigaction = handler;
//     sigemptyset(&sa.sa_mask);
//     sa.sa_flags = SA_SIGINFO;
    
//     if (sigaction(SIGSEGV, &sa, NULL) < 0) {
//         printf ("Cannot setup sigsegv\n");
//         exit(1); 
//     }
// }


int timeval_diff(const struct timeval& t1, const struct timeval& t2)
{
    return (t1.tv_sec - t2.tv_sec) * 1000000 + 
        (t1.tv_usec - t2.tv_usec);
}


struct timeval timeval_incr(const struct timeval& t, int incr_usec)
{
    struct timeval ret;
    ret.tv_usec = t.tv_usec + incr_usec;
    ret.tv_sec = t.tv_sec + ret.tv_usec / 1000000;
    ret.tv_usec = ret.tv_usec % 1000000;
    if (ret.tv_usec < 0) { //Note incr_usec can be negative, and note -11%100 = -11
        ret.tv_sec --;
        ret.tv_usec += 1000000;
    }
    return ret;
}


#ifdef __OS_CYGWIN__
#include <windows.h>
void get_mytime(struct timeval* tm)
{
    //windows API
    __int64 now, freq;
    
    QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
    QueryPerformanceCounter((LARGE_INTEGER*)&now);
    
    tm->tv_sec = now / freq;
    tm->tv_usec = (long)((((double) now / freq) - tm->tv_sec) * 1000000);
 
    //gettimeofday (tm, NULL);
    //printf ("time = %u:%u\n", tm->tv_sec, tm->tv_usec);
    return;
}
#else
void get_mytime(struct timeval* tm) { gettimeofday (tm, NULL); }
#endif
