/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

// Ping algorithm

#ifndef __PA_PING_H__
#define __PA_PING_H__

#include <string>
#include <vector>
using namespace std;

#include "Common.h"
#include "PacketCatcher.h"


// PA_Ping has no idea of how probe packet is crafted and how to parse
// probe reply.  class ProbePacket takes care of this dirty work. Ping
// template requires the following Probe concept:
// 
// class Probe {
// public:
//     void send (uint16 seq);
//     bool parse (char* rawIpPacket, int packet_len, uint16* seq);
// };
// 
// PA_Ping relies on IncomPktMgr_Client to receive raw packet.
// 
// Ping sends up to nprobe packets to the destination (implicitly
// provided by probep).  Ping stops after either reaching nprobe or
// the number of good probes reaches nMaxGoodProbe.  Ping send packet
// by packet; it won't send the next packet until getting the reply or
// the timeout.  The sequence of probe packets starts from
// init_pkt_id, and recycles after reaching max_pkt_id.

template <class ProbePacket>
class PA_Ping
{
public:
    PA_Ping (IncomPktMgr_Client* pkt_clnt, ProbePacket* probep, 
        int nprobe, int nMaxGoodProbe, int timeout_msec, 
        unsigned short init_pkt_id, unsigned short max_pkt_id = 0xffff);
    ~PA_Ping ();
    void run (void);
    string output (void);

private:   
    int wait_for_reply (int timeout_msec, uint16 expected_seq, struct timeval *tm_recv);

    ProbePacket* probep_;
    
    struct Per_Probe_Record
    {
        uint32 from_ip;
        double rtt_msec;
        struct timeval tm_sent;
    };
    
    vector<Per_Probe_Record> records_;
    int nprobe_, nMaxGoodProbe_;
    int timeout_msec_;
    unsigned short init_pkt_id_, max_pkt_id_;

    int ngood_probes;
    IncomPktMgr_Client* pkt_clnt_;
};


#include <iostream>
#include <sstream>
using namespace std;
#include <sys/time.h>
#include <pthread.h>
#include <errno.h>

#include "Common.h"


template <class ProbePacket>
PA_Ping<ProbePacket>::
PA_Ping (IncomPktMgr_Client* pkt_clnt, ProbePacket* probep, int nprobe, int nMaxGoodProbe, int timeout_msec, 
        unsigned short init_pkt_id, unsigned short max_pkt_id)
{
    pkt_clnt_ = pkt_clnt;
    probep_ = probep;
    nprobe_ = nprobe;
    nMaxGoodProbe_ = nMaxGoodProbe;
    timeout_msec_ = timeout_msec;
    init_pkt_id_ = init_pkt_id;
    max_pkt_id_ = max_pkt_id;
    ngood_probes = 0;
}


template <class ProbePacket>
PA_Ping<ProbePacket>::
~PA_Ping ()
{
}

#define MAX_PACKET_LEN 1500

template <class ProbePacket>
void PA_Ping<ProbePacket>::
run (void)
{
    unsigned short current_pkt_id = init_pkt_id_;

    int i;
    for (i = 0;
         i < nprobe_ && ngood_probes < nMaxGoodProbe_;
         i++)
    {
        Per_Probe_Record pr;
        pr.from_ip = 0;
        pr.rtt_msec = -1;
        records_.push_back (pr);
        
        probep_->send (current_pkt_id);
        //gettimeofday (&(records_[i].tm_sent), NULL);
        get_mytime (&(records_[i].tm_sent));

        struct timeval tm_recv;
        //Process the reply
        if (wait_for_reply (timeout_msec_, current_pkt_id, &tm_recv) == 0) {
            ASSERT(i == records_.size() - 1, 
                   "invalid record_q_i_ %d, size %d\n", i, records_.size());
            
            double raw_rtt = timeval_diff (tm_recv, records_[i].tm_sent) / 1000.0;
            records_[i].rtt_msec = (raw_rtt >= 0 ? raw_rtt : 0); // - proc_delay;
            log(INFO, "raw_rtt = %.2f rtt = %.2f", raw_rtt, records_[i].rtt_msec);

            ngood_probes ++;
        }
        
        log(INFO, "Wait done");
        
        current_pkt_id ++;
        if (current_pkt_id >= max_pkt_id_) current_pkt_id = init_pkt_id_;
    }
}


template <class ProbePacket>
int PA_Ping<ProbePacket>::
wait_for_reply (int timeout_msec, uint16 expected_seq, struct timeval *tm_recv)
{
    struct timeval start;
    //gettimeofday (&start, NULL);
    get_mytime (&start);

    while (1) {
        struct timeval now;
        //gettimeofday (&now, NULL);
        get_mytime (&now);
        int rem_timeout_msec = timeout_msec - timeval_diff (now, start) / 1000;
        if (rem_timeout_msec <= 0) {
            return 1; //timeout
        }
        
        char buffer[MAX_PACKET_LEN];
        int packlen = MAX_PACKET_LEN;
        log (INFO, "Call recv with timeout = %d msec", rem_timeout_msec);
        
        if (pkt_clnt_->recv(buffer, &packlen, tm_recv, rem_timeout_msec)) {
            uint16 seq;
            bool ret = probep_->parse (buffer, packlen, &seq);

            if (ret) {
                log(INFO, "parse, seq = %u", seq);
                
                if (seq == expected_seq) {
                    log(INFO, "got reply");
                    return 0;
                }
                else
                    log(INFO, "bad packet id");
            }
            else
                log(INFO, "bad packet");
        }
        else return 1;//timeout
    }
}

template <class ProbePacket>
string PA_Ping<ProbePacket>::
output (void)
{
    ostringstream oss (ostringstream::out);

    struct timeval tm;
    gettimeofday (&tm, NULL);
    
    oss << probep_->info_str () << " good_probes= " << ngood_probes << " finish_time= " << tm.tv_sec << endl;
    
    for (int i = 0; i < records_.size(); i ++) {
        if (records_[i].rtt_msec >= 0)
            oss << i << " " << records_[i].rtt_msec << " ms" << endl;
        else
            oss << i << " *" << endl;
    }

    return oss.str();
}

#endif                                            //__PA_PING_H__
