/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

// IncomPktMgr (Incoming Packet Manager) class suite allows multiple
// threads (pthread) receive incoming packets that arrive at a single
// shared socket.

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>

#include <iostream>
#include <algorithm>
using namespace std;

#include "Common.h"
#include "NetUtil.h"
#include "PacketCatcher.h"

#ifndef INADDR_ANY
#define INADDR_ANY 0
#endif

#define MAX_PACKET_LEN 1500

#ifdef __OS_LINUX__
#include <stropts.h>
#endif


IncomPktMgr::IncomPktMgr (int protocol)
{
    init (&protocol, 1);
}


IncomPktMgr::IncomPktMgr (int* protocols, int nprotocol)
{
    ASSERT (nprotocol >= 1, "");
    init (protocols, nprotocol);
}


void IncomPktMgr::init (int* protocols, int nprotocol)
{
    for (int i = 0; i < MAX_PC_SUBS; i ++)    
        subs_[i] = NULL;

    stop_thread_ = true;
    // Create a mutex
    int res = pthread_mutex_init (&lock_, NULL);
    if (res != 0) throw MyException(str_log(ERROR, "Create a mutex"));

    for (int i = 0; i < nprotocol; i ++) {
        int sock_proto = protocols[i];
#if defined __OS_CYGWIN__ || defined __OS_BSD__
        //cygwin doesn't accept IPPROTO_TCP. You have to listen to IP as a whole.
        sock_proto = (protocols[i] == IPPROTO_TCP) ? IPPROTO_IP: protocols[i];
#elif defined __OS_LINUX__
#else
#error "NO OS"
#endif
        vector<int>::iterator existed = find(protocols_.begin(), protocols_.end(), protocols[i]);

        if (existed != protocols_.end()) {
            log (INFO, "Duplicate sockd. Skip this protocol");
            continue;
        }

        int sockd = socket (PF_INET, SOCK_RAW, sock_proto);
        if (sockd < 0)
            throw MyException(str_log(ERROR, "socket: %s", strerror(errno)));

        struct sockaddr_in addr;
        addr.sin_addr.s_addr = INADDR_ANY;
        addr.sin_family = AF_INET;
        addr.sin_port = htons(0);
        
        if ((bind(sockd, (struct sockaddr *)&addr, sizeof(addr))) < 0)
            throw MyException(str_log(ERROR, "bind"));
        
        log (INFO, "Bind socket to port %u", 0);
        
        int on = 1;
        if(setsockopt (sockd, 0, IP_HDRINCL, (char *)&on, sizeof(on)) < 0)
            throw MyException(str_log(ERROR, "setsockopt"));

        log (INFO, "Create socket of type %d", sock_proto);

        sockds_.push_back (sockd);
        protocols_.push_back (protocols[i]);
    }
}


IncomPktMgr::~IncomPktMgr ()
{
    stop ();
    pthread_mutex_destroy (&lock_);

    for (vector<int>::iterator i = sockds_.begin(); i != sockds_.end(); i ++)
        close (*i);
}


void* _run_thread (void* arg)
{
    IncomPktMgr *me = (IncomPktMgr*)arg;
    me->run_Thread();
    return NULL;
}


void IncomPktMgr::append_Rule (IncomPktMgr_Rule* new_rule)
{
    pthread_mutex_lock (&lock_);
    rule_chain.push_back (new_rule);
    pthread_mutex_unlock (&lock_);
}


void IncomPktMgr::asyn_Run (void)
{
    // Create and execute the listening thread
    stop_thread_ = false;
    int res = pthread_create (&thread_, NULL, _run_thread, this);
    if (res != 0) throw MyException (str_log(ERROR, "Create a thread"));
}


void IncomPktMgr::stop (void)
{
    pthread_mutex_lock (&lock_);
    stop_thread_ = true;
    pthread_mutex_unlock (&lock_);
}


void IncomPktMgr::run_Thread (void)
{
    log (INFO, "Packet Catcher thread starts");

    vector<int> sockds_sort = sockds_;
    sort (sockds_sort.begin (), sockds_sort.end ());
    int max_fd = sockds_sort.back ();
    log (INFO, "max_fd  = %d", max_fd);
    fd_set rset;
    
    while (1)
    {
        pthread_mutex_lock (&lock_);
        if (stop_thread_) {
            pthread_mutex_unlock (&lock_);
            break;
        }            
        pthread_mutex_unlock (&lock_);
        
        FD_ZERO (&rset);
        for (vector<int>::iterator i = sockds_.begin(); i != sockds_.end(); i ++)
            FD_SET (*i, &rset);

        struct timeval timeout;
        timeout.tv_sec = 2; timeout.tv_usec = 0;
        int ret = select (max_fd + 1, &rset, NULL, NULL, &timeout);

        //Have to poll the stop status stop_thread_ again, because it
        //may have changed while this thread was doing select().  If
        //we don't check stop_thread_, stop_thread_ might be already
        //set and the whole classes has been deconstructed.
        pthread_mutex_lock (&lock_);
        if (stop_thread_) {
            pthread_mutex_unlock (&lock_);
            break;
        }            
        pthread_mutex_unlock (&lock_);
        
        if (ret < 0) {
            throw MyException (str_log (ERROR, "select: %s", strerror(errno)));
        }
        else if (ret > 0) {
            log (DUMP, "Select() got something");
            
            for (vector<int>::iterator i = sockds_.begin(); i != sockds_.end(); i ++) {
                if (FD_ISSET (*i, &rset)) {
                    struct timeval tm;
                    char buffer[MAX_PACKET_LEN];
                    //blocking recv

                    int data_len = recv (*i, buffer, MAX_PACKET_LEN, 0);

                    if (data_len > 0) {
#if defined __OS_CYGWIN__
                        //gettimeofday (&tm, NULL);
                        get_mytime (&tm);
#elif defined __OS_BSD__
                        gettimeofday (&tm, NULL);
#elif defined __OS_LINUX__
                        //Get kernel time stamp
                        if (ioctl (*i, SIOCGSTAMP, &tm)) {
                            throw MyException(str_log(ERROR, "ioctl"));
                        }
#endif

                        log (DUMP, "Got a packet");
                        handle_Packet (buffer, data_len, tm);
                    }
                    else
                        log (WARN, "recv returns %d", data_len);
                }
            }
        }
        //else timeout
    }
}


void IncomPktMgr::addClient (IncomPktMgr_Client* subs, int id)
{
    ASSERT (id >= 0 && id < MAX_PC_SUBS, "id %d", id);
    pthread_mutex_lock (&lock_);
    ASSERT (subs_[id] == NULL, 
        "Client with id %d already added. You forgot to remove?", id);
    subs_[id] = subs;
    pthread_mutex_unlock (&lock_);
    log (DUMP, "add client id=%d", id);
}


void IncomPktMgr::removeClient (int id)
{
    ASSERT (id >= 0 && id < MAX_PC_SUBS, "id %d", id);
    pthread_mutex_lock (&lock_);
    ASSERT (subs_[id] != NULL, "Client with id %d not exists.", id);
    subs_[id] = NULL;
    pthread_mutex_unlock (&lock_);
    log (DUMP, "remove client id=%d", id);
}


void IncomPktMgr::handle_Packet (char* packet, int packlen, struct timeval tm)
{
    log (DUMP, "handle_Packet");

    pthread_mutex_lock (&lock_);
      
    for (list<IncomPktMgr_Rule*>::iterator r = rule_chain.begin();
        r != rule_chain.end(); r ++) 
    {
        int packetThrdId = (*r)->pkt_To_ClientId (packet, packlen);
        //printf ("id = %d\n", packetThrdId);        

        if (packetThrdId >= 0 && packetThrdId < MAX_PC_SUBS && subs_[packetThrdId]) {
            log (DUMP, "notify client id=%d", packetThrdId);
            IncomPktMgr_Packet t;
            t.pkt = new char[packlen];
            memcpy (t.pkt, packet, packlen);
            t.pkt_len = packlen;
            t.recv_tm = tm;
            subs_[packetThrdId]->pkt_buffer.push_back (t);

            pthread_cond_signal (&(subs_[packetThrdId]->cond_wait_));
            break;
        }
        //else { printf ("not delivered\n"); }
    }

    pthread_mutex_unlock (&lock_);
}


IncomPktMgr_Client::IncomPktMgr_Client (void)
{
    int res;
    res = pthread_mutex_init (&lock_, NULL);
    if (res != 0)        
        throw MyException(str_log(ERROR, "Create a mutex"));
    res = pthread_cond_init (&cond_wait_, NULL);
    if (res != 0)        
        throw MyException(str_log(ERROR, "Create a thread condition"));
}


IncomPktMgr_Client::~IncomPktMgr_Client ()
{
    pthread_mutex_destroy (&lock_);
    pthread_cond_destroy (&cond_wait_);
}


bool IncomPktMgr_Client::recv (char* pkt, int* pkt_len, struct timeval* recv_tm, int timeout_msec)
{
    pthread_mutex_lock (&lock_);

    log (DUMP, "recv");
    
    if (pkt_buffer.size() == 0) {
        //wait for an arrival
        struct timeval now, timeout_tmp;
        struct timespec timeout;

        gettimeofday (&now, NULL);
        //get_mytime (&now);
        timeout_tmp = timeval_incr (now, timeout_msec * 1000);
        timeout.tv_sec  = timeout_tmp.tv_sec;
        timeout.tv_nsec = timeout_tmp.tv_usec * 1000;
        
        log(DUMP, "waiting for packet arrival, timeout %d msec, thrdid = %u", timeout_msec, pthread_self());
        //log(INFO, "waiting for packet arrival, timeout %u:%u", timeout.tv_sec, timeout.tv_nsec);
        
        //release lock and wait on cond_wait
        //NOTE: the 3rd arg should be absolute time
        int res = pthread_cond_timedwait (&cond_wait_, &lock_, &timeout);

        log(DUMP, "pthread_cond_timedwait: %d, thrdid = %u", res, pthread_self());
        if (res != 0) {
            if (res == ETIMEDOUT)
            {
                log(DUMP, "timeout");
            }
            else {
                throw MyException(str_log(ERROR, "pthread_cond_timedwait %s",
                                          strerror(res)));
            }
        }
    }
    
    bool ret = false;
    if (pkt_buffer.size() > 0) {
        if (pkt_buffer.front().pkt_len > *pkt_len) {
            //Overflowing packets
            log(WARN, "space too small to hold the packet");
            ret = false;
        }
        else {
            *pkt_len = pkt_buffer.front().pkt_len;
            memcpy (pkt, pkt_buffer.front().pkt, pkt_buffer.front().pkt_len);
            delete [] pkt_buffer.front().pkt;
            if (recv_tm != NULL) *recv_tm = pkt_buffer.front().recv_tm;
            ret = true;
        }
        pkt_buffer.pop_front ();
    }

    pthread_mutex_unlock (&lock_);
    return ret;
}


