/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fstream>
using namespace std;

#include "Common.h"
#include "NetUtil.h"
#include "PA_Ping.h"
#include "PacketCatcher.h"
#include "PP_TcpAck.h"
#include "PP_IcmpEcho.h"
#include "TF_MtBatch.h"

enum WorkMode { WM_SINGLE, WM_MTBATCH }; // WM_MTINTER };
enum PingProtocol { PING_PROTO_TCP_ACK, PING_PROTO_ICMP_ECHO };

struct Option {
    WorkMode wm;
    PingProtocol proto;
    int count, good_count;
    struct in_addr saddr;
    unsigned short dport;
    int debug;
    struct in_addr daddr;

    struct Mt_Option {
        ifstream dest_list;
        uint16 port;
        int con_thrds;
        int intra_thrd_bits;
        ~Mt_Option() { if (dest_list.is_open()) dest_list.close(); }        
    };
    Mt_Option mt;
};

struct Option opt;

void usage (void)
{
    printf ("Ping a network destination or a list of destinations. Version %s. Author: Zheng Zhang.\n", VERSION);
    printf ("\n");
    printf ("Usage: iPing [Options] [Target]\n");
    printf ("\n");
    printf ("Target: the destination IP or dns name. Target should be empty if \"-F\" option is used.\n");
//    printf ("\n");
    printf ("Options:\n");
    printf ("\nBasic\n---------\n");
    printf ("  -R PROTOCOL      send probing packet in protocol PROTOCOL. PROTOCOL = tcpack, icmp. Default is tcpack.\n");
    printf ("  -S srcIP         use srcIP as source IP.\n");
    printf ("  -p destPort      use destPort as destination TCP port number. Default is 33000.\n");
    printf ("  -c COUNT         send COUNT number of probes (serially). Default is 3.\n");
    printf ("  -g good_count    send probes (serially) until having collected good_count number of good probes.\n");
    printf ("\nConcurrency\n---------\n");
    printf ("  -C THREADS       use THREADS number of parallel threads. Default is 10.\n");
    printf ("  -F dest_list     probe the destinations in file dest_list.\n");
    printf ("\nMisc\n---------\n");
    printf ("  -d debug_level   set debug level. 1: error, 2: warn, 3: info, 4: dump. Default is 2.\n");
    printf ("  -h               display this help.\n");
}

void init_option (void)
{
    opt.debug = WARN;
    opt.proto = PING_PROTO_TCP_ACK;
    opt.dport = 33000;
    opt.daddr.s_addr = 0;
    opt.saddr.s_addr = 0;
    opt.count = 3;
    opt.good_count = -1;
    opt.wm = WM_SINGLE;
    opt.mt.con_thrds = 10;
}

void parse_cmd (int argc, char *argv[])
{
    int c;
  
    init_option ();
    
    while ((c = getopt(argc, argv, "R:p:S:c:d:C:F:g:h")) > 0) {
        switch(c) {
        case 'R':
            if (! strcmp("tcp_ack", optarg)) {
                opt.proto = PING_PROTO_TCP_ACK;
            }
            else if (! strcmp("icmp", optarg)) {
                opt.proto = PING_PROTO_ICMP_ECHO;
            }
            else {
                printf ("Invalid protocol name in \"-R %s\"\n", optarg);
                exit (1);
            }
            break;
        case 'p':
            opt.dport = (unsigned short)atoi(optarg);
            break;
        case 'S':
            opt.saddr.s_addr = inet_addr(optarg);
            if (opt.saddr.s_addr == INADDR_NONE) {
                printf ("Bad source address: %s\n", optarg);
                exit (1);
            }
            break;
        case 'c':
            opt.count = atoi(optarg);
            break;
        case 'g':
            opt.good_count = atoi(optarg);
            break;
        case 'C':
            opt.mt.con_thrds = atoi(optarg);
            if (opt.mt.con_thrds > 128) {
                printf("-C concurrent threads too large. Must not exceed 128\n");
                exit (1);
            }
            break;
        case 'F':
            opt.mt.dest_list.open (optarg, ifstream::in);
            if (! opt.mt.dest_list.is_open()) {
                printf("Cannot open file %s\n", optarg);
                exit(1);
            }
            if (opt.wm != WM_SINGLE) {
                printf("Other work modes such as -I cannot co-exists with -F");
                exit(1);
            }
            opt.wm = WM_MTBATCH;
            break;
//         case 'I':
//             if (opt.wm != WM_SINGLE) {
//                 printf("Other work modes such as -F cannot co-exists with -I");
//                 exit(1);
//             }
//             opt.wm = WM_MTINTER;
//             opt.mt.port = atoi(optarg);
//             break;
        case 'd':
            opt.debug = atoi(optarg);
            break;
        case 'h':
            usage(); exit (0);
        default:
            usage(); exit (1);
        }
    }

    if (opt.good_count <= 0) opt.good_count = opt.count;
    
    argc = argc - optind + 1;
    argv = argv + optind - 1;

    if (opt.wm == WM_SINGLE) {
        if (argc != 2) {
            usage();
            exit(1);
        }
        
        struct hostent *he;  
        if((he = (struct hostent *)gethostbyname(argv[1])) == NULL) {  /* get the host info */
            printf("Cannot resolve the input address\n");
            exit(0);
        }
        opt.daddr = *((struct in_addr *)he->h_addr);
    } else { //WM_MTBATCH
        if (argc != 1) {
            usage ();
            exit (1);
        }
    }

    set_log_level (opt.debug);
}

typedef PA_Ping<PP_TcpAck> TcpPing;
typedef PA_Ping<PP_IcmpEcho> IcmpPing;

template <class base, class trait_base, class DummyP>
class _TF : public base {
public:
    _TF (void);
    bool process_single_task (int thrd_ith, const string& line, string* output);            
private:
    IncomPktMgr pc;
    //PP_TcpAck dummy_tcp;
    //IncomPktMgr_Rule_MT<PP_TcpAck> pc_rule;
    DummyP dummy;
    IncomPktMgr_Rule_MT<DummyP> pc_rule;
};


template <class base, class trait_base, class DummyP>
_TF<base, trait_base, DummyP>::_TF (void)
    :base (opt.mt.con_thrds, trait_base::get_2nd_contruct_arg()),
     pc (opt.proto == PING_PROTO_TCP_ACK ? IPPROTO_TCP : IPPROTO_ICMP),
     dummy (0, opt.dport),
     pc_rule (&dummy, opt.mt.intra_thrd_bits)
{
    pc.append_Rule (&pc_rule);
    pc.asyn_Run ();
}


template <class base, class trait_base, class DummyP>
bool
_TF<base, trait_base, DummyP>::process_single_task (int thrd_ith, const string& line, string* output)
{
    //printf ("thrd_ith = %d\n", thrd_ith);
    struct hostent *he;
    if((he = (struct hostent *)gethostbyname(line.c_str())) == NULL) {
        *output = string ("Cannot resolve the input address\n");
        return true;
    }
    struct in_addr daddr = *((struct in_addr *)he->h_addr);

    IncomPktMgr_Client pkt_clnt;

    unsigned short init_seq = ((unsigned short) thrd_ith) << opt.mt.intra_thrd_bits;
    unsigned short max_seq = ((unsigned short) (thrd_ith + 1)) << opt.mt.intra_thrd_bits;
    //printf ("init %u max %u intra_bits %d\n", init_seq, max_seq, opt.mt.intra_thrd_bits);

    if (opt.proto == PING_PROTO_TCP_ACK) {
        PP_TcpAck tcp (daddr.s_addr, opt.dport);
        TcpPing ping (&pkt_clnt, &tcp, opt.count, opt.good_count, 2000, init_seq, max_seq);
        
        pc.addClient (&pkt_clnt, thrd_ith);
        ping.run ();
        pc.removeClient (thrd_ith);
        
        *output = ping.output ();
    }
    else if (opt.proto == PING_PROTO_ICMP_ECHO) {
        PP_IcmpEcho icmp (daddr.s_addr, opt.dport);
        IcmpPing ping (&pkt_clnt, &icmp, opt.count, opt.good_count, 2000, init_seq, max_seq);
        
        pc.addClient (&pkt_clnt, thrd_ith);
        ping.run ();
        pc.removeClient (thrd_ith);
        
        *output = ping.output ();
    }
    else { ASSERT (0, "What else?"); }
    
    return true;    
}

class TF_MtFileBatch_trait {
public:
    static ifstream* get_2nd_contruct_arg () { return &(opt.mt.dest_list); }
};

// class TF_MtSockInteractive_trait {
// public:
//     static uint16 get_2nd_contruct_arg () { return opt.mt.port; }
// };

typedef _TF<TF_MtFileBatch, TF_MtFileBatch_trait, PP_TcpAck> _TFB_tcp;
typedef _TF<TF_MtFileBatch, TF_MtFileBatch_trait, PP_IcmpEcho> _TFB_icmp;
// typedef _TF<TF_MtSockInteractive, TF_MtSockInteractive_trait> _TFI;

int main_single (void)
{
    srand (getpid());
    unsigned short init_seq = rand() % time(NULL);

    //TODO: need to clean up the code duplication. We need to use smart pointer.
    if (opt.proto == PING_PROTO_TCP_ACK) {
        IncomPktMgr pc (IPPROTO_TCP); 
        IncomPktMgr_Client pkt_clnt;
        pc.addClient (&pkt_clnt, 0);
        IncomPktMgr_Rule_Default default_rule;
        pc.append_Rule (&default_rule);
        
        PP_TcpAck tcp (opt.daddr.s_addr, opt.dport);
        TcpPing ping (&pkt_clnt, &tcp, opt.count, opt.good_count, 2000, init_seq);

        pc.asyn_Run ();
        ping.run ();
        
        cout << ping.output();
    }
    else if (opt.proto == PING_PROTO_ICMP_ECHO) {
        IncomPktMgr pc (IPPROTO_ICMP);
        IncomPktMgr_Client pkt_clnt;
        pc.addClient (&pkt_clnt, 0);
        IncomPktMgr_Rule_Default default_rule;
        pc.append_Rule (&default_rule);
        
        PP_IcmpEcho icmp (opt.daddr.s_addr, opt.dport);        
        IcmpPing ping (&pkt_clnt, &icmp, opt.count, opt.good_count, 2000, init_seq);
        
        pc.asyn_Run ();
        ping.run ();
        
        cout << ping.output();
    }
    else { ASSERT (0, "What else?"); }
    
    return 0;
}


int main_mtbatch (void)
{
    int thrd_id_bits = 0;
    for (unsigned int t = 1; opt.mt.con_thrds > t; t = t << 1, thrd_id_bits ++);
    opt.mt.intra_thrd_bits = 16 - thrd_id_bits;
    ASSERT (opt.mt.intra_thrd_bits >= 2,
        "opt.mt.intra_thrd_bits is too small, probably because con_thrds is too large");
    
    if (opt.proto == PING_PROTO_TCP_ACK) {
        _TFB_tcp tf;
        tf.run ();
    } else {
        _TFB_icmp tf;
        tf.run ();
    }
    
    return 0;
}


int main (int argc, char** argv)
{
    init_option ();
    parse_cmd (argc, argv);

    try {
        if (opt.wm == WM_SINGLE)
            return main_single ();
        else
            return main_mtbatch ();
    }
    catch (MyException e) {
    }
}

