/*************************************************************/
/* iTraceroute (part of iSPY) Project                        */
/* Copyright (c) 2008, Zheng Zhang, Purdue University        */
/*************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fstream>
using namespace std;

#include "Common.h"
#include "NetUtil.h"
#include "TextUtil.h"
#include "PA_Traceroute.h"
#include "PacketCatcher.h"
#include "PP_ttlTcp.h"
#include "TF_MtBatch.h"
#include "patricia.h"

enum WorkMode { WM_SINGLE, WM_MTBATCH };

struct Option {
    WorkMode wm;
    struct in_addr saddr;
    struct in_addr daddr;
    unsigned short dport;
    int count, good_count;
    int abort_hops;
    uint8 start_ttl, max_ttl;
    patricia_tree_t* ip2AS_mapping;
//    int timeout_msec;
    int stop_enter_AS;
    int debug;

    struct Mt_Option {
        ifstream dest_list;
        uint16 port;
        int con_thrds;
        int intra_thrd_bits;
        ~Mt_Option() { if (dest_list.is_open()) dest_list.close(); }        
    };
    Mt_Option mt;

    //void spawn_local (struct Option& local_opt);
    /*
    ~Option() 
    //was, but triggers segfault
    {
        if (ip2AS_mapping) { 
            log (INFO, "Destroy patricia"); 
            Destroy_Patricia (ip2AS_mapping, 0); 
        } 
    }
    */
};

/*
void Option::spawn_local (struct Option& local_opt)
{
    local_opt.wm = wm;
    local_opt.saddr = saddr;
    local_opt.daddr = daddr;
    local_opt.dport = dport;
    local_opt.count = count;
    local_opt.abort_hops = abort_hops;
    local_opt.start_ttl = start_ttl;
    local_opt.max_ttl = max_ttl;
    local_opt.ip2AS_mapping = ip2AS_mapping;
    local_opt.stop_enter_AS = stop_enter_AS;
}
*/

struct Option opt;

void usage (void)
{
    printf ("Traceroute a network destination or a list of destinations. Version %s. Author: Zheng Zhang.\n", VERSION);
    printf ("\n");
    printf ("Usage: iTraceroute [Options] [Target]\n");
    printf ("\n");
    printf ("Target: the destination IP or dns name. Target should be empty if \"-F\" option is used.\n");
//    printf ("\n");
    printf ("Options:\n");
    printf ("\nBasic\n---------\n");
    printf ("  -S srcIP         use srcIP as source IP.\n");
    printf ("  -p destPort      use destPort as destination TCP port number. Default is 33000.\n");
    printf ("  -c COUNT         send COUNT number of probes (serially) per hop. Default is 3.\n");
    printf ("  -g good_count    send probes (serially) until having collected good_count number of good probes.\n");
    printf ("  -f start_ttl     start with TTL start_ttl. Default is 1.\n");
    printf ("  -m max_ttl       end with TTL max_ttl. Default is 30.\n");
    printf ("  -a abort_hops    abort after abort_hops number of consecutive hop fails. Default is 3.\n");
    printf ("\nConcurrency\n---------\n");
    printf ("  -C THREADS       use THREADS number of parallel threads. Default is 10.\n");
    printf ("  -F dest_list     probe the destinations in file dest_list.\n");
    printf ("\nIP-to-AS mapping\n---------\n");
    printf ("  -M mapping_file  read IP-to-AS mapping from file mapping_file.\n");
    printf ("  -A stop_AS            stop probing if reaching AS stop_AS.\n");
    printf ("\nMisc\n---------\n");
    printf ("  -d debug_level   set debug level. 1: error, 2: warn, 3: info, 4: dump. Default is 2.\n");
    printf ("  -h               display this help.\n");
}

void init_option (struct Option& opt)
{
    opt.debug = WARN;
    opt.dport = 33000;
    opt.daddr.s_addr = 0;
    opt.saddr.s_addr = 0;
    opt.count = -1;
    opt.good_count = -1;
    opt.abort_hops = -1;
    opt.start_ttl = 0;
    opt.max_ttl = 0;
    opt.ip2AS_mapping = NULL;
    opt.stop_enter_AS = -1;
    opt.wm = WM_SINGLE;
    opt.mt.con_thrds = 10;
}

void parse_cmd (int argc, char *argv[], bool local, struct Option& opt)
{
    if (! local)
        set_log_level (opt.debug);

    optind = 0; //reset getopt

    const char* opt_str = local ? 
        "A:" : 
        "p:S:c:g:d:M:f:m:C:F:I:A:a:";

    //for (int i = 0; i < argc; i++) printf ("INSIDE: ARGV= %s#\n", argv[i]);

    int c;
    while ((c = getopt(argc, argv, opt_str)) > 0) {
        switch(c) {
        case 'p':
            opt.dport = (unsigned short)atoi(optarg);
            break;
        case 'S':
            opt.saddr.s_addr = inet_addr(optarg);
            if (opt.saddr.s_addr == INADDR_NONE) {
                printf ("Bad source address: %s\n", optarg);
                exit (1);
            }
            break;
        case 'c':
            opt.count = atoi(optarg);
            break;
        case 'g':
            opt.good_count = atoi(optarg);
            break;
        case 'a':
            opt.abort_hops = atoi(optarg);
            break;
        case 'f':
            opt.start_ttl = atoi(optarg);
            break;
        case 'm':
            opt.max_ttl = atoi(optarg);
            break;
        case 'M':
            opt.ip2AS_mapping = Ip2AsMapping_tree_init (optarg);
            break;
        case 'A':
            opt.stop_enter_AS = atoi(optarg);
            break;
        case 'd':
            opt.debug = atoi(optarg);
            set_log_level (opt.debug);
            break;
        case 'C':
            opt.mt.con_thrds = atoi(optarg);
            if (opt.mt.con_thrds > 128) {
                printf("-C concurrent threads too large. Must not exceed 128\n");
                exit (1);
            }
            break;
        case 'F':
            opt.mt.dest_list.open (optarg, ifstream::in);
            if (! opt.mt.dest_list.is_open()) {
                printf("Cannot open file %s\n", optarg);
                exit(1);
            }
            opt.wm = WM_MTBATCH;
            break;
        default:
            usage(); exit (1);
        }
    }

    //if (opt.good_count <= 0) opt.good_count = opt.count;
    
    argc = argc - optind + 1;
    argv = argv + optind - 1;

    if (opt.wm == WM_SINGLE || local) {
        if (argc != 2) {
            usage();
            exit(1);
        }
        
        struct hostent *he;  
        if((he = (struct hostent *)gethostbyname(argv[1])) == NULL) {  /* get the host info */
            printf("Cannot resolve the input address\n");
            exit(0);
        }
        opt.daddr = *((struct in_addr *)he->h_addr);
    } else { //WM_MTBATCH
        if (argc != 1) {
            usage ();
            exit (1);
        }
    }
}

int proto[] = { IPPROTO_ICMP, IPPROTO_TCP };
typedef PA_Traceroute<PP_ttlTcp> TcpTracert;
    
template <class base, class trait_base>
class _TF : public base {
public:
    _TF (void);
    bool process_single_task (int thrd_ith, const string& line, string* output);            
private:
    IncomPktMgr pc;
    PP_ttlTcp dummy_tcp;
    IncomPktMgr_Rule_MT<PP_ttlTcp> pc_rule;
};


template <class base, class trait_base>
_TF<base, trait_base>::_TF (void)
    :base (opt.mt.con_thrds, trait_base::get_2nd_contruct_arg()),
      pc (proto, 2),
      dummy_tcp (opt.daddr.s_addr, opt.dport),
      pc_rule (&dummy_tcp, opt.mt.intra_thrd_bits)
{
    pc.append_Rule (&pc_rule);
    pc.asyn_Run ();
}


template <class base, class trait_base>
bool
_TF<base, trait_base>::process_single_task (int thrd_ith, const string& line, string* output)
{
    //printf ("thrd_ith = %d\n", thrd_ith);
/*Was
    struct hostent *he;
    if((he = (struct hostent *)gethostbyname(line.c_str())) == NULL) {
        *output = string ("Cannot resolve the input address\n");
        return true;
    }
    struct in_addr daddr = *((struct in_addr *)he->h_addr);
*/
    struct Option local_opt;
    init_option (local_opt);

    //Split line.c_str() into argc, argv;
    vector<string> argv_v;
    TextUtil::split (line, argv_v, string(" "));
    
    int argc = argv_v.size() + 1;
    //printf ("#ARGC= %d\n", argc);
    typedef char* char_ptr;
    char_ptr* argv = new char_ptr [argc];
    argv[0] = strdup ("single_task");
    for (int i = 1; i < argc; i ++) {
        argv[i] = strdup (argv_v[i - 1].c_str());
    }
    //for (int i = 0; i < argc; i ++)         printf ("#ARGV=%s\n", argv[i]);
    
    parse_cmd (argc, argv, true, local_opt);
    
    for (int i = 0; i < argc; i ++) free (argv[i]);
    delete[] argv;

    IncomPktMgr_Client pkt_clnt;

    PP_ttlTcp tcp (local_opt.daddr.s_addr, opt.dport);
    unsigned short init_seq = ((unsigned short) thrd_ith) << opt.mt.intra_thrd_bits;
    unsigned short max_seq = ((unsigned short) (thrd_ith + 1)) << opt.mt.intra_thrd_bits;
    //printf ("init %u max %u intra_bits %d\n", init_seq, max_seq, opt.mt.intra_thrd_bits);

    TcpTracert tracert (&pkt_clnt, &tcp, init_seq, max_seq);

    if (opt.count > 0) tracert.addOption_nProbe_per_Hop (opt.count);
    if (opt.good_count > 0) tracert.addOption_nMaxGoodProbe_per_Hop (opt.good_count);
    if (opt.abort_hops > 0) tracert.addOption_Consecu_Noresp_Abort (opt.abort_hops);
    if (opt.start_ttl > 0) tracert.addOption_StartTTL (opt.start_ttl);
    if (opt.max_ttl > 0) tracert.addOption_MaxTTL (opt.max_ttl);
    if (opt.ip2AS_mapping) tracert.addOption_IP2AS_Mapping (opt.ip2AS_mapping);
    if (local_opt.stop_enter_AS > 0) tracert.addOption_Enter_Stop_AS (local_opt.stop_enter_AS);
    if (opt.stop_enter_AS > 0) tracert.addOption_Enter_Stop_AS (opt.stop_enter_AS);
                        
    pc.addClient (&pkt_clnt, thrd_ith);
    try {
        tracert.run ();
    }
    catch (MyException e) {
        fprintf (stderr, "Exception with %s. Skip\n", line.c_str());
    }
    pc.removeClient (thrd_ith);
    
    *output = tracert.output ();
    return true;    
}


class TF_MtFileBatch_trait {
public:
    static ifstream* get_2nd_contruct_arg () { return &(opt.mt.dest_list); }
};

//class TF_MtSockInteractive_trait {
//public:
//    static uint16 get_2nd_contruct_arg () { return opt.mt.port; }
//};

typedef _TF<TF_MtFileBatch, TF_MtFileBatch_trait> _TFB;
//typedef _TF<TF_MtSockInteractive, TF_MtSockInteractive_trait> _TFI;


int main_single (void)
{
    IncomPktMgr pc (proto, 2);    
    IncomPktMgr_Client pkt_clnt;
    pc.addClient (&pkt_clnt, 0);
    IncomPktMgr_Rule_Default default_rule;
    pc.append_Rule (&default_rule);

    PP_ttlTcp tcp (opt.daddr.s_addr, opt.dport);

    srand (getpid());
    unsigned short init_seq = rand() % time(NULL);
    TcpTracert tracert (&pkt_clnt, &tcp, init_seq);

    if (opt.count > 0) tracert.addOption_nProbe_per_Hop (opt.count);
    if (opt.good_count > 0) tracert.addOption_nMaxGoodProbe_per_Hop (opt.good_count);
    if (opt.abort_hops > 0) tracert.addOption_Consecu_Noresp_Abort (opt.abort_hops);
    if (opt.start_ttl > 0) tracert.addOption_StartTTL (opt.start_ttl);
    if (opt.max_ttl > 0) tracert.addOption_MaxTTL (opt.max_ttl);
    if (opt.ip2AS_mapping) tracert.addOption_IP2AS_Mapping (opt.ip2AS_mapping);
    if (opt.stop_enter_AS > 0) tracert.addOption_Enter_Stop_AS (opt.stop_enter_AS);
    
    pc.asyn_Run ();
    tracert.run ();

    cout << tracert.output();
    return 0;
}


int main_mtbatch (void)
{
    int thrd_id_bits = 0;
    for (unsigned int t = 1; opt.mt.con_thrds > t; t = t << 1, thrd_id_bits ++);
    opt.mt.intra_thrd_bits = 16 - thrd_id_bits;
    ASSERT (opt.mt.intra_thrd_bits >= 2,
        "opt.mt.intra_thrd_bits is too small, probably because con_thrds is too large");
    
//  if (opt.wm == WM_MTBATCH) {    
        _TFB tf;
        tf.run ();
/*    } else {
        _TFI tf;
        tf.run ();
    }
*/
}


int main (int argc, char** argv)
{
//    printf ("#ARGC= %d\n", argc);
//    for (int i = 0; i < argc; i ++) printf ("argv=%s\n", argv[i]);

    init_option (opt);
    parse_cmd (argc, argv, false, opt);
    try {
        if (opt.wm == WM_SINGLE)
            return main_single ();
        else
            return main_mtbatch ();
    }
    catch (MyException e) {
    }
}

